"""Shared pytest fixtures for the aiodhcpwatcher test suite."""

from collections.abc import Iterator

import pytest
from blockbuster import BlockBuster, blockbuster_ctx


@pytest.fixture(autouse=True)
def blockbuster() -> Iterator[BlockBuster]:
    """Fail the test if aiodhcpwatcher makes a blocking call in the event loop."""
    with blockbuster_ctx("aiodhcpwatcher") as bb:
        yield bb



import asyncio
import logging
import os
import time
from datetime import datetime, timedelta, timezone
from functools import partial
from unittest.mock import MagicMock, patch

import pytest
import pytest_asyncio
from scapy import (
    arch,  # noqa: F401
    interfaces,
)
from scapy.error import Scapy_Exception
from scapy.layers.l2 import Ether
from scapy.packet import Packet

from aiodhcpwatcher import (
    AUTO_RECOVER_TIME,
    DHCPRequest,
    async_init,
    async_start,
    make_packet_handler,
)

utcnow = partial(datetime.now, timezone.utc)
_MONOTONIC_RESOLUTION = time.get_clock_info("monotonic").resolution

logging.basicConfig(level=logging.DEBUG)


def async_fire_time_changed(utc_datetime: datetime) -> None:
    timestamp = utc_datetime.timestamp()
    loop = asyncio.get_running_loop()
    for task in list(loop._scheduled):  # type: ignore[attr-defined]
        if not isinstance(task, asyncio.TimerHandle):
            continue
        if task.cancelled():
            continue

        mock_seconds_into_future = timestamp - time.time()
        future_seconds = task.when() - (loop.time() + _MONOTONIC_RESOLUTION)

        if mock_seconds_into_future >= future_seconds:
            task._run()
            task.cancel()


# connect b8:b7:f1:6d:b5:33 192.168.210.56
RAW_DHCP_REQUEST = (
    b"\xff\xff\xff\xff\xff\xff\xb8\xb7\xf1m\xb53\x08\x00E\x00\x01P\x06E"
    b"\x00\x00\xff\x11\xb4X\x00\x00\x00\x00\xff\xff\xff\xff\x00D\x00C\x01<"
    b"\x0b\x14\x01\x01\x06\x00jmjV\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb8\xb7\xf1m\xb53\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00c\x82Sc5\x01\x039\x02\x05\xdc2\x04\xc0\xa8\xd286"
    b"\x04\xc0\xa8\xd0\x017\x04\x01\x03\x1c\x06\x0c\x07connect\xff\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)

# connec\xdb b8:b7:f1:6d:b5:33 192.168.210.56
DHCP_REQUEST_BAD_UTF8 = (
    b"\xff\xff\xff\xff\xff\xff\xb8\xb7\xf1m\xb53\x08\x00E\x00\x01P\x06E"
    b"\x00\x00\xff\x11\xb4X\x00\x00\x00\x00\xff\xff\xff\xff\x00D\x00C\x01<"
    b"\x0b\x14\x01\x01\x06\x00jmjV\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb8\xb7\xf1m\xb53\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00c\x82Sc5\x01\x039\x02\x05\xdc2\x04\xc0\xa8\xd286"
    b"\x04\xc0\xa8\xd0\x017\x04\x01\x03\x1c\x06\x0c\x07connec\xab\xff\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)
# ó b8:b7:f1:6d:b5:33 192.168.210.56
DHCP_REQUEST_IDNA = (
    b"\xff\xff\xff\xff\xff\xff\xb8\xb7\xf1m\xb53\x08\x00E\x00\x01P\x06E"
    b"\x00\x00\xff\x11\xb4X\x00\x00\x00\x00\xff\xff\xff\xff\x00D\x00C\x01<"
    b"\x0b\x14\x01\x01\x06\x00jmjV\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xb8\xb7\xf1m\xb53\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00c\x82Sc5\x01\x039\x02\x05\xdc2\x04\xc0\xa8\xd286"
    b"\x04\xc0\xa8\xd0\x017\x04\x01\x03\x1c\x06\x0c\x07xn--kda\xff\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)

# iRobot-AE9EC12DD3B04885BCBFA36AFB01E1CC 50:14:79:03:85:2c 192.168.1.120
RAW_DHCP_RENEWAL = (
    b"\x00\x15\x5d\x8e\xed\x02\x50\x14\x79\x03\x85\x2c\x08\x00\x45\x00"
    b"\x01\x8e\x51\xd2\x40\x00\x40\x11\x63\xa1\xc0\xa8\x01\x78\xc0\xa8"
    b"\x01\x23\x00\x44\x00\x43\x01\x7a\x12\x09\x01\x01\x06\x00\xd4\xea"
    b"\xb2\xfd\xff\xff\x00\x00\xc0\xa8\x01\x78\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x50\x14\x79\x03\x85\x2c\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x63\x82\x53\x63\x35\x01\x03\x39\x02\x05"
    b"\xdc\x3c\x45\x64\x68\x63\x70\x63\x64\x2d\x35\x2e\x32\x2e\x31\x30"
    b"\x3a\x4c\x69\x6e\x75\x78\x2d\x33\x2e\x31\x38\x2e\x37\x31\x3a\x61"
    b"\x72\x6d\x76\x37\x6c\x3a\x51\x75\x61\x6c\x63\x6f\x6d\x6d\x20\x54"
    b"\x65\x63\x68\x6e\x6f\x6c\x6f\x67\x69\x65\x73\x2c\x20\x49\x6e\x63"
    b"\x20\x41\x50\x51\x38\x30\x30\x39\x0c\x27\x69\x52\x6f\x62\x6f\x74"
    b"\x2d\x41\x45\x39\x45\x43\x31\x32\x44\x44\x33\x42\x30\x34\x38\x38"
    b"\x35\x42\x43\x42\x46\x41\x33\x36\x41\x46\x42\x30\x31\x45\x31\x43"
    b"\x43\x37\x08\x01\x21\x03\x06\x1c\x33\x3a\x3b\xff"
)

# <no hostname> 60:6b:bd:59:e4:b4 192.168.107.151
RAW_DHCP_REQUEST_WITHOUT_HOSTNAME = (
    b"\xff\xff\xff\xff\xff\xff\x60\x6b\xbd\x59\xe4\xb4\x08\x00\x45\x00"
    b"\x02\x40\x00\x00\x00\x00\x40\x11\x78\xae\x00\x00\x00\x00\xff\xff"
    b"\xff\xff\x00\x44\x00\x43\x02\x2c\x02\x04\x01\x01\x06\x00\xff\x92"
    b"\x7e\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x60\x6b\xbd\x59\xe4\xb4\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x63\x82\x53\x63\x35\x01\x03\x3d\x07\x01"
    b"\x60\x6b\xbd\x59\xe4\xb4\x3c\x25\x75\x64\x68\x63\x70\x20\x31\x2e"
    b"\x31\x34\x2e\x33\x2d\x56\x44\x20\x4c\x69\x6e\x75\x78\x20\x56\x44"
    b"\x4c\x69\x6e\x75\x78\x2e\x31\x2e\x32\x2e\x31\x2e\x78\x32\x04\xc0"
    b"\xa8\x6b\x97\x36\x04\xc0\xa8\x6b\x01\x37\x07\x01\x03\x06\x0c\x0f"
    b"\x1c\x2a\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
)


async def _write_test_packets_to_pipe(w: int) -> None:
    for test_packet in (
        RAW_DHCP_REQUEST_WITHOUT_HOSTNAME,
        RAW_DHCP_REQUEST,
        RAW_DHCP_RENEWAL,
        RAW_DHCP_REQUEST_WITHOUT_HOSTNAME,
        DHCP_REQUEST_BAD_UTF8,
        DHCP_REQUEST_IDNA,
    ):
        os.write(w, test_packet)
        for _ in range(3):
            await asyncio.sleep(0)
        os.write(w, b"garbage")
        for _ in range(3):
            await asyncio.sleep(0)


class MockIface:
    index: int = 0

    def __init__(self) -> None:
        MockIface.index = MockIface.index + 1
        self.index: int = MockIface.index


class MockSocket:

    def __init__(self, reader: int, exc: type[Exception] | None = None) -> None:
        self._fileno = reader
        if reader != -1:
            # The real listen socket is set non-blocking in _make_listen_socket,
            # so the mock reads from a non-blocking fd too. This keeps os.read in
            # _on_data off the event loop's blocking radar (see blockbuster).
            os.set_blocking(reader, False)
        self.iface = MockIface()
        self.close = MagicMock()
        self.buffer = b""
        self.exc = exc

    def recv(self) -> Packet:
        if self.exc:
            raise self.exc
        raw = os.read(self._fileno, 1000000)
        try:
            packet = Ether(raw)
        except Exception:
            packet = Packet(raw)
        return packet

    def fileno(self) -> int:
        return self._fileno


@pytest_asyncio.fixture(autouse=True, scope="session")
async def _init_scapy():
    await async_init()


@pytest.mark.asyncio
async def test_start_stop():
    """Test start and stop."""
    (await async_start(lambda data: None))()


@pytest.mark.asyncio
async def test_watcher():
    """Test mocking a dhcp packet to the watcher."""
    requests: list[DHCPRequest] = []

    def _handle_dhcp_packet(data: DHCPRequest) -> None:
        requests.append(data)

    r, w = os.pipe()

    mock_socket = MockSocket(r)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        stop = await async_start(_handle_dhcp_packet)
        await _write_test_packets_to_pipe(w)

        stop()

    os.close(r)
    os.close(w)
    assert requests == [
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connect",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.1.120",
            hostname="iRobot-AE9EC12DD3B04885BCBFA36AFB01E1CC",
            mac_address="50:14:79:03:85:2c",
        ),
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connec�",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="ó",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
    ]


@pytest.mark.asyncio
async def test_watcher_fatal_exception(caplog: pytest.LogCaptureFixture) -> None:
    """Test mocking a dhcp packet to the watcher."""
    requests: list[DHCPRequest] = []

    def _handle_dhcp_packet(data: DHCPRequest) -> None:
        requests.append(data)

    r, w = os.pipe()

    mock_socket = MockSocket(r, ValueError)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        stop = await async_start(_handle_dhcp_packet)
        await _write_test_packets_to_pipe(w)

        stop()

    os.close(r)
    os.close(w)
    assert requests == []
    assert "Fatal error while processing dhcp packet" in caplog.text


@pytest.mark.asyncio
async def test_watcher_temp_exception(caplog: pytest.LogCaptureFixture) -> None:
    """Test mocking a dhcp packet to the watcher."""
    requests: list[DHCPRequest] = []

    def _handle_dhcp_packet(data: DHCPRequest) -> None:
        requests.append(data)

    r, w = os.pipe()
    mock_socket = MockSocket(r, OSError)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        stop = await async_start(_handle_dhcp_packet)
        await _write_test_packets_to_pipe(w)
    os.close(r)
    os.close(w)
    assert requests == []
    assert "Error while processing dhcp packet" in caplog.text

    r, w = os.pipe()
    mock_socket = MockSocket(r)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):

        async_fire_time_changed(utcnow() + timedelta(seconds=AUTO_RECOVER_TIME))
        await asyncio.sleep(0.1)

        await _write_test_packets_to_pipe(w)

        stop()

    os.close(r)
    os.close(w)
    assert requests == [
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connect",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.1.120",
            hostname="iRobot-AE9EC12DD3B04885BCBFA36AFB01E1CC",
            mac_address="50:14:79:03:85:2c",
        ),
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connec�",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="ó",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
    ]


@pytest.mark.asyncio
async def test_watcher_if_indexes(caplog: pytest.LogCaptureFixture) -> None:
    """Test mocking a dhcp packet to the watcher."""
    requests: list[DHCPRequest] = []

    def _handle_dhcp_packet(data: DHCPRequest) -> None:
        requests.append(data)

    r, w = os.pipe()
    mock_socket = MockSocket(r)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        stop = await async_start(_handle_dhcp_packet, if_indexes=[1, 2])
        await _write_test_packets_to_pipe(w)
        stop()

    os.close(r)
    os.close(w)
    assert requests == [
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connect",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.1.120",
            hostname="iRobot-AE9EC12DD3B04885BCBFA36AFB01E1CC",
            mac_address="50:14:79:03:85:2c",
        ),
        DHCPRequest(
            ip_address="192.168.107.151", hostname="", mac_address="60:6b:bd:59:e4:b4"
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="connec�",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
        DHCPRequest(
            ip_address="192.168.210.56",
            hostname="ó",
            mac_address="b8:b7:f1:6d:b5:33",
        ),
    ]


@pytest.mark.asyncio
async def test_watcher_stop_after_temp_exception(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test mocking a dhcp packet to the watcher."""
    requests: list[DHCPRequest] = []

    def _handle_dhcp_packet(data: DHCPRequest) -> None:
        requests.append(data)

    r, w = os.pipe()

    mock_socket = MockSocket(r, OSError)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        stop = await async_start(_handle_dhcp_packet)
        await _write_test_packets_to_pipe(w)

    os.close(r)
    os.close(w)
    assert requests == []
    assert "Error while processing dhcp packet" in caplog.text
    stop()

    r, w = os.pipe()
    mock_socket = MockSocket(r)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):

        async_fire_time_changed(utcnow() + timedelta(seconds=30))
        await asyncio.sleep(0)
        await _write_test_packets_to_pipe(w)

        stop()

    os.close(r)
    os.close(w)
    assert requests == []


@pytest.mark.asyncio
async def test_setup_fails_broken_filtering(caplog: pytest.LogCaptureFixture) -> None:
    """Test that the setup fails when filtering is broken."""
    caplog.set_level(logging.DEBUG)
    with patch(
        "scapy.arch.common.compile_filter",
        side_effect=Scapy_Exception,
    ):
        (await async_start(lambda data: None))()
    assert (
        "Cannot watch for dhcp packets without a functional packet filter"
        in caplog.text
    )


@pytest.mark.asyncio
async def test_setup_fails_as_root(caplog: pytest.LogCaptureFixture) -> None:
    """Test that the setup fails as root."""
    with (
        patch("os.geteuid", return_value=0),
        patch("scapy.arch.common.compile_filter"),
        patch.object(
            interfaces,
            "resolve_iface",
            side_effect=Scapy_Exception,
        ),
    ):
        (await async_start(lambda data: None))()
    assert "Cannot watch for dhcp packets" in caplog.text


@pytest.mark.asyncio
async def test_setup_fails_as_non_root(caplog: pytest.LogCaptureFixture) -> None:
    """Test that the setup fails as root."""
    caplog.set_level(logging.DEBUG)
    with (
        patch("os.geteuid", return_value=10),
        patch("scapy.arch.common.compile_filter"),
        patch.object(
            interfaces,
            "resolve_iface",
            side_effect=Scapy_Exception,
        ),
    ):
        (await async_start(lambda data: None))()
    assert "Cannot watch for dhcp packets without root or CAP_NET_RAW" in caplog.text


@pytest.mark.asyncio
async def test_permission_denied_to_add_reader(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test permission denied to add reader."""
    loop = asyncio.get_running_loop()
    with (
        patch.object(loop, "add_reader", side_effect=PermissionError),
        patch("scapy.arch.common.compile_filter"),
        patch.object(
            interfaces,
            "resolve_iface",
        ),
    ):
        (await async_start(lambda data: None))()

    assert "Permission denied to watch for dhcp packets" in caplog.text


def test_hostname_idna_unicode_error_does_not_crash_handler() -> None:
    """
    A crafted DHCP hostname must not crash the packet handler.

    On Python < 3.13 the ``idna`` codec raises a bare ``UnicodeError`` (e.g.
    "label empty or too long") instead of ``UnicodeDecodeError`` for malformed
    hostnames. The handler previously caught only ``UnicodeDecodeError``, so an
    untrusted DHCP packet from the LAN could raise an uncaught exception inside
    the asyncio reader callback. This test simulates that codec behaviour in a
    version-independent way and asserts the handler swallows it and falls back
    to a lossy utf-8 decode.
    """
    from scapy.layers.dhcp import DHCP
    from scapy.layers.inet import IP
    from scapy.layers.l2 import Ether

    class _IdnaUnicodeErrorBytes(bytes):
        """bytes whose idna decode raises a bare UnicodeError (pre-3.13)."""

        def decode(self, encoding: str = "utf-8", errors: str = "strict") -> str:
            if encoding == "idna":
                raise UnicodeError("label empty or too long")
            return super().decode(encoding, errors)

    requests: list[DHCPRequest] = []
    handler = make_packet_handler(requests.append)

    packet = (
        Ether(src="b8:b7:f1:6d:b5:33")
        / IP(src="192.168.210.56")
        / DHCP(
            options=[
                ("message-type", 3),
                ("hostname", _IdnaUnicodeErrorBytes(b"xn--bad")),
                "end",
            ]
        )
    )

    # Must not raise, even though idna decode raises a bare UnicodeError.
    handler(packet)

    assert len(requests) == 1
    assert requests[0].hostname == "xn--bad"
    assert requests[0].ip_address == "192.168.210.56"
    assert requests[0].mac_address == "b8:b7:f1:6d:b5:33"


def test_handler_no_ether_layer_does_not_crash() -> None:
    """
    A packet without an Ethernet layer must not crash the handler.

    ``make_packet_handler`` is a public, exported entrypoint. Not every link
    layer carries an ``Ether`` header: Linux cooked captures (the ``any``
    pseudo-device uses an ``SLL`` header), PPP, tun, and some cellular/VPN
    interfaces have no Ethernet framing, so ``packet.getlayer(Ether)`` returns
    ``None``. The handler unconditionally read ``getlayer(Ether).src``, raising
    ``AttributeError: 'NoneType' object has no attribute 'src'`` on every DHCP
    packet seen on such an interface. With no MAC available the packet should
    simply be skipped, not crash the asyncio reader callback.
    """
    from scapy.layers.dhcp import BOOTP, DHCP
    from scapy.layers.inet import IP, UDP

    requests: list[DHCPRequest] = []
    handler = make_packet_handler(requests.append)

    # IP present (requested_addr absent → renewal path), but no Ether layer.
    packet = (
        IP(src="192.168.1.50")
        / UDP(sport=68, dport=67)
        / BOOTP()
        / DHCP(options=[("message-type", 3), "end"])
    )
    assert packet.getlayer(Ether) is None

    # Must not raise; with no MAC the request is dropped.
    handler(packet)
    assert requests == []


def test_handler_no_ip_layer_does_not_crash() -> None:
    """
    A packet without an IP layer must not crash the handler.

    When a DHCP request carries no ``requested_addr`` option (e.g. a renewal),
    the handler falls back to ``packet.getlayer(IP).src``. If the packet has no
    IP layer that read raised ``AttributeError``. Such a packet has no usable
    client address and should be skipped rather than crash the handler.
    """
    from scapy.layers.dhcp import BOOTP, DHCP
    from scapy.layers.inet import IP

    requests: list[DHCPRequest] = []
    handler = make_packet_handler(requests.append)

    # Ether present, but no IP layer and no requested_addr → IP.src fallback.
    packet = (
        Ether(src="aa:bb:cc:dd:ee:ff")
        / BOOTP()
        / DHCP(options=[("message-type", 3), "end"])
    )
    assert packet.getlayer(IP) is None

    # Must not raise; with no IP the request is dropped.
    handler(packet)
    assert requests == []


def test_all_exports_are_importable() -> None:
    """
    Every name advertised in ``__all__`` must be a real module attribute.

    Regression test: ``__all__`` listed ``"start"`` long after the
    module-level ``start`` function was renamed to ``async_start``, so
    ``from aiodhcpwatcher import *`` raised ``AttributeError: module
    'aiodhcpwatcher' has no attribute 'start'``.
    """
    import aiodhcpwatcher

    for name in aiodhcpwatcher.__all__:
        assert hasattr(
            aiodhcpwatcher, name
        ), f"{name!r} is declared in __all__ but not defined in the module"


def test_async_start_is_exported() -> None:
    """``async_start`` is the documented entrypoint and must be exported."""
    import aiodhcpwatcher

    assert "async_start" in aiodhcpwatcher.__all__


@pytest.mark.asyncio
async def test_invalid_file_descriptor(caplog: pytest.LogCaptureFixture) -> None:
    """
    Test an invalid (-1) socket file descriptor is handled gracefully.

    On platforms such as Windows scapy's listen socket does not expose a
    selectable file descriptor, so fileno() returns -1 and add_reader would
    raise ValueError. The watcher must not crash.
    """
    mock_socket = MockSocket(-1)
    with (
        patch(
            "aiodhcpwatcher.AIODHCPWatcher._make_listen_socket",
            return_value=mock_socket,
        ),
        patch("aiodhcpwatcher.AIODHCPWatcher._verify_working_pcap"),
    ):
        (await async_start(lambda data: None))()

    assert "valid file descriptor" in caplog.text


@pytest.mark.asyncio
async def test_add_reader_not_supported(caplog: pytest.LogCaptureFixture) -> None:
    """
    Test add_reader being unsupported is handled gracefully.

    The Windows Proactor event loop does not implement add_reader, which
    raises NotImplementedError. The watcher must degrade gracefully.
    """
    loop = asyncio.get_running_loop()
    with (
        patch.object(loop, "add_reader", side_effect=NotImplementedError),
        patch("scapy.arch.common.compile_filter"),
        patch.object(interfaces, "resolve_iface"),
    ):
        (await async_start(lambda data: None))()

    assert "Cannot watch for dhcp packets" in caplog.text



__version__ = "1.2.7"

import asyncio
import logging
import os
import socket
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any

from scapy.config import conf
from scapy.error import Scapy_Exception

if TYPE_CHECKING:
    from scapy.packet import Packet

_LOGGER = logging.getLogger(__name__)

FILTER = "udp and (port 67 or 68)"
DHCP_REQUEST = 3
AUTO_RECOVER_TIME = 30


@dataclass(slots=True)
class DHCPRequest:
    """Class to hold a DHCP request."""

    ip_address: str
    hostname: str
    mac_address: str


def make_packet_handler(
    callback: Callable[[DHCPRequest], None],
) -> Callable[["Packet"], None]:
    """Create a packet handler."""
    # Local import because importing from scapy has side effects such as opening
    # sockets
    from scapy.layers.dhcp import DHCP  # pylint: disable=import-outside-toplevel
    from scapy.layers.inet import IP  # pylint: disable=import-outside-toplevel
    from scapy.layers.l2 import Ether  # pylint: disable=import-outside-toplevel

    def _handle_dhcp_packet(packet: "Packet") -> None:
        """Process a dhcp packet."""
        if not (dhcp_packet := packet.getlayer(DHCP)):
            return

        options: Iterable[tuple[str, int | bytes | None] | str] = dhcp_packet.options

        options_dict: dict[str, int | bytes | None] = {}
        for option in options:
            if type(option) is tuple and len(option) > 1:
                options_dict[option[0]] = option[1]
            if option == "end":
                break

        if options_dict.get("message-type") != DHCP_REQUEST:
            # Not a DHCP request
            return

        ip_address: str | None = options_dict.get("requested_addr")  # type: ignore[assignment]
        if not ip_address:
            # No requested_addr option (e.g. a renewal): fall back to the source
            # IP. Packets captured on link layers without IP framing have no IP
            # layer, so guard against getlayer() returning None.
            ip_layer = packet.getlayer(IP)
            ip_address = ip_layer.src if ip_layer is not None else None

        hostname = ""
        if (hostname_bytes := options_dict.get("hostname")) and isinstance(
            hostname_bytes, bytes
        ):
            try:
                # The standard uses idna encoding for hostnames, but some clients
                # do not follow the standard and use utf-8 instead.
                hostname = hostname_bytes.decode("idna")
            except UnicodeError:
                # On Python < 3.13 the idna codec can raise a bare UnicodeError
                # (e.g. "label empty or too long") rather than UnicodeDecodeError
                # for crafted hostnames. Catch the base class so an untrusted
                # DHCP packet cannot crash the packet handler.
                hostname = hostname_bytes.decode("utf-8", errors="replace")

        # Not every link layer carries an Ethernet header (e.g. Linux cooked
        # capture / SLL, PPP, tun), so getlayer(Ether) can be None; without a
        # MAC there is nothing useful to report.
        ether_layer = packet.getlayer(Ether)
        mac_address: str | None = ether_layer.src if ether_layer is not None else None

        if ip_address is not None and mac_address is not None:
            callback(DHCPRequest(ip_address, hostname, mac_address))

    return _handle_dhcp_packet


class AIODHCPWatcher:
    """Class to watch dhcp requests."""

    _init_scapy_done = False

    def __init__(self, callback: Callable[[DHCPRequest], None]) -> None:
        """Initialize watcher."""
        self._loop = asyncio.get_running_loop()
        self._socks: list[tuple[int, socket.socket, int]] = []
        self._callback = callback
        self._shutdown: bool = False
        self._restart_timer: asyncio.TimerHandle | None = None
        self._restart_task: asyncio.Task[None] | None = None

    def restart_soon(self) -> None:
        """Restart the watcher soon."""
        if not self._restart_timer:
            _LOGGER.debug("Restarting watcher in %s seconds", AUTO_RECOVER_TIME)
            self._restart_timer = self._loop.call_later(
                AUTO_RECOVER_TIME, self._execute_restart
            )

    def _clear_restart_task(self, task: asyncio.Task[None]) -> None:
        """Clear the restart task."""
        self._restart_task = None

    def _execute_restart(self) -> None:
        """Execute the restart."""
        self._restart_timer = None
        if not self._shutdown:
            _LOGGER.debug("Restarting watcher")
            self._restart_task = self._loop.create_task(self.async_start())
            self._restart_task.add_done_callback(self._clear_restart_task)

    def shutdown(self) -> None:
        """Shutdown the watcher."""
        self._shutdown = True
        self.stop()

    def stop(self) -> None:
        """Stop watching for DHCP packets."""
        if self._restart_timer:
            self._restart_timer.cancel()
            self._restart_timer = None
        if self._restart_task:
            self._restart_task.cancel()
            self._restart_task = None
        if self._socks is not None:
            for _, sock, fileno in self._socks:
                self._loop.remove_reader(fileno)
                sock.close()
            self._socks = []

    def _start(
        self, if_indexes: Iterable[int] | None = None
    ) -> Callable[["Packet"], None] | None:
        """Start watching for dhcp packets."""
        _init_scapy()
        # disable scapy promiscuous mode as we do not need it
        conf.sniff_promisc = 0

        try:
            self._verify_working_pcap(FILTER)
        except (Scapy_Exception, ImportError) as ex:
            _LOGGER.error(
                "Cannot watch for dhcp packets without a functional packet filter: %s",
                ex,
            )
            return None

        for if_index in set(if_indexes) if if_indexes else [None]:
            try:
                if sock := self._make_listen_socket(FILTER, if_index):
                    if if_index is None:
                        if_index = sock.iface.index
                    self._socks.append((if_index, sock, sock.fileno()))
            except (Scapy_Exception, OSError) as ex:
                if os.geteuid() == 0:
                    _LOGGER.error("Cannot watch for dhcp packets: %s", ex)
                else:
                    _LOGGER.debug(
                        "Cannot watch for dhcp packets without root or CAP_NET_RAW: %s",
                        ex,
                    )
                return None

        return make_packet_handler(self._callback)

    async def async_start(self, if_indexes: Iterable[int] | None = None) -> None:
        """Start watching for dhcp packets."""
        if self._shutdown:
            _LOGGER.debug("Not starting watcher because it is shutdown")
            return
        if not (
            _handle_dhcp_packet := await self._loop.run_in_executor(
                None, self._start, if_indexes
            )
        ):
            return
        if self._shutdown:  # may change during the executor call
            _LOGGER.debug("Not starting watcher because it is shutdown after init")  # type: ignore[unreachable]
            return
        for if_index, sock, fileno in list(self._socks):
            if fileno == -1:
                # On some platforms (notably Windows) scapy's listen socket does
                # not expose a selectable file descriptor, so fileno() returns -1
                # and add_reader cannot watch it. This platform is unsupported.
                _LOGGER.error(
                    "Cannot watch for dhcp packets on %s because the socket does "
                    "not have a valid file descriptor; this platform is likely "
                    "unsupported",
                    if_index,
                )
                sock.close()
                self._socks.remove((if_index, sock, fileno))
                continue
            try:
                self._loop.add_reader(
                    fileno, partial(self._on_data, _handle_dhcp_packet, sock)
                )
                _LOGGER.debug("Started watching for dhcp packets on %s", if_index)
            except PermissionError as ex:
                _LOGGER.error(
                    "Permission denied to watch for dhcp packets on %s: %s",
                    if_index,
                    ex,
                )
                sock.close()
                self._socks.remove((if_index, sock, fileno))
            except (NotImplementedError, ValueError) as ex:
                # Some event loops (e.g. the Windows Proactor loop) do not
                # implement add_reader, and some sockets report an invalid file
                # descriptor. Degrade gracefully instead of crashing.
                _LOGGER.error(
                    "Cannot watch for dhcp packets on %s: %s",
                    if_index,
                    ex,
                )
                sock.close()
                self._socks.remove((if_index, sock, fileno))
        if len(self._socks) == 0:
            _LOGGER.debug("Not starting watcher because no readers added")

    def _on_data(
        self, handle_dhcp_packet: Callable[["Packet"], None], sock: Any
    ) -> None:
        """Handle data from the socket."""
        try:
            data = sock.recv()
        except (BlockingIOError, InterruptedError):
            return
        except OSError as ex:
            _LOGGER.error("Error while processing dhcp packet: %s", ex)
            self.stop()
            self.restart_soon()
            return
        except BaseException as ex:  # pylint: disable=broad-except
            _LOGGER.exception("Fatal error while processing dhcp packet: %s", ex)
            self.shutdown()
            return

        if data:
            handle_dhcp_packet(data)

    def _make_listen_socket(self, cap_filter: str, if_index: int | None = None) -> Any:
        """Get a nonblocking listen socket."""
        from scapy.data import ETH_P_ALL  # pylint: disable=import-outside-toplevel
        from scapy.interfaces import (  # pylint: disable=import-outside-toplevel
            dev_from_index,
            resolve_iface,
        )

        iface = dev_from_index(if_index) if if_index and if_index > -1 else conf.iface
        sock = resolve_iface(iface).l2listen()(
            type=ETH_P_ALL, iface=iface, filter=cap_filter
        )
        if hasattr(sock, "set_nonblock"):
            # Not all classes have set_nonblock so we have to call fcntl directly
            # in the event its not implemented
            sock.set_nonblock(True)
        elif hasattr(sock, "pcap_fd"):
            sock.pcap_fd.setnonblock(True)
        else:
            import fcntl  # pylint: disable=import-outside-toplevel

            fcntl.fcntl(sock.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

        return sock

    def _verify_working_pcap(self, cap_filter: str) -> None:
        """
        Verify we can create a packet filter.

        If we cannot create a filter we will be listening for
        all traffic which is too intensive.
        """
        # Local import because importing from scapy has side effects such as opening
        # sockets
        from scapy.arch.common import (  # pylint: disable=import-outside-toplevel
            compile_filter,
        )

        compile_filter(cap_filter)


async def async_start(
    callback: Callable[[DHCPRequest], None], if_indexes: Iterable[int] | None = None
) -> Callable[[], None]:
    """Listen for DHCP requests."""
    watcher = AIODHCPWatcher(callback)
    await watcher.async_start(if_indexes)
    return watcher.shutdown


async def async_init() -> None:
    """Init scapy in the executor since it blocks for a bit."""
    await asyncio.get_running_loop().run_in_executor(None, _init_scapy)


def _init_scapy() -> None:
    """Init scapy in the executor since it blocks for a bit."""
    # Local import because importing from scapy has side effects such as opening
    # sockets
    # We must import l2 before testing the filter or it will fail

    #
    # Importing scapy.sendrecv will cause a scapy resync which will
    # import scapy.arch.read_routes which will import scapy.sendrecv
    #
    # We avoid this circular import by importing arch above to ensure
    # the module is loaded and avoid the problem
    #
    if AIODHCPWatcher._init_scapy_done:
        return
    from scapy import arch  # pylint: disable=import-outside-toplevel # noqa: F401
    from scapy.arch.common import (  # pylint: disable=import-outside-toplevel # noqa: F401
        compile_filter,
    )
    from scapy.layers import (
        l2,  # pylint: disable=import-outside-toplevel # noqa: F401
    )
    from scapy.layers.dhcp import (
        DHCP,  # pylint: disable=import-outside-toplevel # noqa: F401
    )
    from scapy.layers.inet import (
        IP,  # pylint: disable=import-outside-toplevel # noqa: F401
    )
    from scapy.layers.l2 import (
        Ether,  # pylint: disable=import-outside-toplevel # noqa: F401
    )

    AIODHCPWatcher._init_scapy_done = True


__all__ = ["DHCPRequest", "async_init", "async_start", "make_packet_handler"]
